package com.gct.aoid.algorithm;

import com.alibaba.fastjson.JSON;
import org.springframework.stereotype.Component;
import org.springframework.util.StringUtils;

import java.math.BigDecimal;
import java.util.*;
import java.util.stream.Collectors;

/**
 * class name: CalDigramProduction
 * 计算单张功图的产液量
 *
 * @author lloyd
 * @version 1.0
 * @since 2021/1/28 下午5:14
 */
@Component
public class CalDiagramProductionCopy {
    public final static ThreadLocal<String> javaScript = new ThreadLocal<>();

    static public class CalProductionResult {
        public Double rcyl1;
        public Double rcyl;
        public Double rcsl;
        public Double rcql;
        public Double bx;
        public Double llpl;
        public Double sjcc;
        public Double yxcc;
        public Double zcc;
        public Double ycc;
        public Double ccss;
        public Double lsss;
        public Double gyss;
        public Double l_StrokeRatio;
        public String remark;
        public Double dym;

        public void appendRemark(String remark) {
            if (Objects.nonNull(this.remark) && !this.remark.isEmpty())
                this.remark += "\n";
            this.remark += remark;
        }

        private double round(double val, int pecison) {
            return BigDecimal.valueOf(val).setScale(pecison, BigDecimal.ROUND_HALF_UP).doubleValue();
        }

        public void formatted() {
            if (Objects.nonNull(rcyl)) rcyl = round(rcyl, 1);
            if (Objects.nonNull(rcyl1)) rcyl1 = round(rcyl1, 1);
            if (Objects.nonNull(rcql)) rcql = round(rcql, 1);
            if (Objects.nonNull(rcsl)) rcsl = round(rcsl, 1);
            if (Objects.nonNull(bx)) bx = round(bx, 2);
            if (Objects.nonNull(llpl)) llpl = round(llpl, 1);
            if (Objects.nonNull(sjcc)) sjcc = round(sjcc, 1);
            if (Objects.nonNull(yxcc)) yxcc = round(yxcc, 1);
            if (Objects.nonNull(zcc)) zcc = round(zcc, 1);
            if (Objects.nonNull(ycc)) ycc = round(ycc, 1);
            if (Objects.nonNull(ccss)) ccss = round(ccss, 2);
            if (Objects.nonNull(lsss)) lsss = round(lsss, 2);
            if (Objects.nonNull(gyss)) gyss = round(gyss, 2);
            if (Objects.nonNull(l_StrokeRatio)) l_StrokeRatio = round(l_StrokeRatio, 2);
            if (Objects.nonNull(dym)) dym = round(dym, 1);
        }
    }

    //region 有效冲程


    /**
     * 计算有效冲程
     *
     * @param bgt
     * @param groupPointNum 必须是奇数
     * @param model
     * @return
     */
    public Double calSpe(Double[][] bgt, int groupPointNum, double model) {

        BGTParam param = new BGTParam();
        Double[][] rBGT = rBGT(bgt, param);
        return calSpe(bgt, rBGT, param, groupPointNum, model);
    }

    private final int STANDARD_WIDTH = 1920;
    private final int STANDARD_HEIGHT = 1080;
    private final Double STANDARD_PADDING = 0.1;

    /**
     * 标准化功图
     *
     * @param bgt
     * @return
     */
    public Double[][] standard(Double[][] bgt, BGTParam param) {
        ArrayList<Double[]> st = new ArrayList<>(bgt.length);
        double xOffset = STANDARD_WIDTH * STANDARD_PADDING;
        double yOffset = STANDARD_HEIGHT * STANDARD_PADDING;
        double widthRegion = STANDARD_WIDTH - xOffset * 2;
        double heightRegion = STANDARD_HEIGHT - yOffset * 2;

        for (int i = 0; i < bgt.length; i++) {
            double x = (bgt[i][0] - param.minS) / (param.maxS - param.minS) * widthRegion;
            double y = (bgt[i][1] - param.minF) / (param.maxF - param.minF) * heightRegion;
            st.add(new Double[]{x + xOffset, y + yOffset});
        }

        return st.toArray(new Double[][]{});
    }

    @Deprecated
    private Integer getHead(List<Integer> range) throws NoSuchFieldError {
        if (Objects.isNull(range)) throw new NoSuchFieldError();
        int minStart = -1, minValCount = range.size();
        switch (minValCount) {
            case 0:
                throw new NoSuchFieldError("不存在的min范围值");
            case 1:
                minStart = range.get(0);
                break;
            default:
                int i = minValCount - 1;
                for (; i > 0; i--) {
                    if (range.get(i) != range.get(i - 1)) {
                        minStart = range.get(i);
                        break;
                    }
                }
                if (i == 0) minStart = range.get(i);
                break;
        }

        return minStart;
    }

    /**
     * 线性差值处理 根据冲程
     *
     * @param param
     * @param line
     * @return
     */
    public ArrayList<Double[]> denseBySF(BGTParam param, List<Double[]> line) {
        double precisionX = (param.maxS - param.minS) / 100;
        double precisionY = (param.maxF - param.minF) / 100;
        ArrayList<Double[]> denseLine = new ArrayList<>();
        for (int i = 0; i < line.size(); i++) {
            Double[] p1 = line.get(i);
            denseLine.add(p1);
            if (i + 1 < line.size()) {
                Double[] p2 = line.get(i + 1);
                double dx = Math.abs(p2[0] - p1[0]) / precisionX;
                double dy = Math.abs(p2[1] - p1[1]) / precisionY;
                int dI = (int) Math.ceil(Math.sqrt(dx * dx + dy * dy));
                if (dI > 1) {//密度不够，精度不够需要插值，线性插值即可
                    for (int j = 1; j < dI; j++)
                        denseLine.add(
                                new Double[]{
                                        (p2[0] - p1[0]) / dI * j + p1[0],
                                        (p2[1] - p1[1]) / dI * j + p1[1]
                                });
                }
            }
        }
        return denseLine;
    }

    /**
     * 计算线的曲率
     *
     * @param line
     * @param groupNum 曲率单元 最小是3 必须是奇数
     * @return
     */
    public LinkedList<Double[]> getLineCurvatures(Double[][] line, int groupNum) {
        LinkedList<Double[]> curvatures = new LinkedList<>();
        LinkedList<Double[]> groupPoints = new LinkedList<>();
        int start = groupNum / 2;
        for (int i = 0; i < groupNum; i++) groupPoints.add(line[i]);
        double v = calCurvature(groupPoints);
        curvatures.add(new Double[]{line[start][0], v});
        for (int index = start + 1; index < line.length - start; index++) {
            groupPoints.remove(0);
            groupPoints.add(line[index + start]);
            v = calCurvature(groupPoints);
            curvatures.add(new Double[]{line[index][0], v});
        }
        return curvatures;

    }

    private boolean comparePoint(Double[] p1, Double[] p2) {
        if (p1.length == p2.length) {
            for (int i = 0; i < p1.length; i++)
                if (p1[i].compareTo(p2[i]) != 0)
                    return false;
            return true;
        }
        return false;
    }

    /**
     * 用下冲程的精确填充上冲程首尾点<br/>
     * 优化计算曲率时的起始边界点误差
     *
     * @param groupNum
     * @param upDense
     * @param downDense
     * @return 交换入参 upDense downDense 返回的即是下冲程填充直
     */
    private List<Double[]> fillHeadTail(int groupNum, List<Double[]> upDense, List<Double[]> downDense) {
        LinkedList<Double[]> upHeadFill = new LinkedList<>();
        for (int i = downDense.size() - 1; i < downDense.size(); i--) {
            if (comparePoint(downDense.get(i), upDense.get(0))) continue;
            for (int j = 0; j < groupNum / 2; j++) {
                upHeadFill.addFirst(downDense.get(j + i));
            }
            break;
        }
        upHeadFill.addAll(upDense);
        for (int i = 0; i < downDense.size(); i++) {
            if (comparePoint(downDense.get(i), upDense.get(upDense.size() - 1))) continue;
            for (int j = 0; j < groupNum / 2; j++) {
                upHeadFill.add(downDense.get(j + i));
            }
            break;
        }
        return upHeadFill;
    }

    private TreeMap<Integer, Double[]> turnPointIndex(LinkedList<Double[]> curvatures, List<Double[]> dense) {


        double temp = -1;//曲率必然是>=0的
        int tempIndex = -1;
        TreeMap<Integer, Double[]> res = new TreeMap<>();
        curvatures.addFirst(new Double[]{null, 0d});
        curvatures.add(new Double[]{null, 0d});
        for (int i = 1; i < curvatures.size() - 1; i++) {
            Double[] cur = curvatures.get(i);
            if (curvatures.get(i - 1)[1] < cur[1] && cur[1] > curvatures.get(i + 1)[1])
                res.put(i - 1, dense.get(i - 1));
        }
        return res;
    }

    public Double calSpeSupper(Double[][] bgt, BGTParam param) {

        Double[][] ps = new Melkman(bgt).convexHullPoint();//此算法运行后，这里输出的功图变成了逆时针的了
//        javaScript.set(javaScript.get().replace("${newOutLine}", JSON.toJSONString(ps)));


        //region 精确定位上下冲程
        int minStart = -1, minEnd = -1, minLastPosition = -1;//注意此时的功图还是逆时针的
        int maxStart = -1, maxEnd = -1, maxLastPosition = -1;
        for (int i = 0; i < ps.length; i++) {
            if (ps[i][0].compareTo(param.minS) == 0) {
                if (minLastPosition == -1) {
                    minStart = i;
                    minEnd = i;
                } else if (minLastPosition + 1 != i) {
                    minStart = i;
                    minEnd = minLastPosition;
                }
                minLastPosition = i;
            }
            if (ps[i][0].compareTo(param.maxS) == 0) {
                if (maxLastPosition == -1) {
                    maxStart = i;
                    maxEnd = i;
                } else if (maxLastPosition + 1 != i) {
                    maxStart = i;
                    maxEnd = maxLastPosition;
                }
                maxLastPosition = i;
            }
        }

        if (minStart == -1 || minEnd == -1 || maxStart == -1 || maxEnd == -1) {
            throw new RuntimeException();
        }

        LinkedList<Double[]> up = new LinkedList<>();
        LinkedList<Double[]> down = new LinkedList<>();


        //minEnd ---- maxStart 逆序 为上冲程
        int temp = minEnd;
        while (temp != maxStart) {
            up.add(ps[temp]);
            temp--;
            if (temp < 0) temp += ps.length;
        }
        if (maxStart != minEnd) up.add(ps[maxStart]);//补

        //maxEnd ---- minStart 逆序 为下冲程
        temp = maxEnd;
        while (temp != minStart) {
            down.add(ps[temp]);
            temp--;
            if (temp < 0) temp += ps.length;
        }
        if (maxEnd != minStart) down.add(ps[minStart]);//补

        javaScript.set(javaScript.get().replace("${up}", JSON.toJSONString(up)));
        javaScript.set(javaScript.get().replace("${down}", JSON.toJSONString(down)));
        List<Double[]> upE = curvatureOptimize(up);
        List<Double[]> downE = curvatureOptimize(down);
        javaScript.set(javaScript.get().replace("${upTurn}", JSON.toJSONString(up)));
        javaScript.set(javaScript.get().replace("${downTurn}", JSON.toJSONString(down)));
        //endregion


        //region 密集化差值。消除轻微波动以达到降噪效果
        ArrayList<Double[]> upDense = denseBySF(param, up);
        ArrayList<Double[]> downDense = denseBySF(param, down);

        javaScript.set(javaScript.get().replace("${upDense}", JSON.toJSONString(upDense)));
        javaScript.set(javaScript.get().replace("${downDense}", JSON.toJSONString(downDense)));
        //endregion

        //region 计算曲率
        int groupNum = 3;
        List<Double[]> upDenseFill = fillHeadTail(groupNum, up, downDense);
        List<Double[]> downDenseFill = fillHeadTail(groupNum, down, upDense);
        javaScript.set(javaScript.get().replace("${upDense}", JSON.toJSONString(upDenseFill)));
        javaScript.set(javaScript.get().replace("${downDense}", JSON.toJSONString(downDenseFill)));
        LinkedList<Double[]> upLineCurvatures = getLineCurvatures(upDenseFill.toArray(new Double[][]{}), groupNum);
        LinkedList<Double[]> downLineCurvatures = getLineCurvatures(downDenseFill.toArray(new Double[][]{}), groupNum);
        javaScript.set(javaScript.get().replace("${upCurvatures}", JSON.toJSONString(upLineCurvatures)));
        javaScript.set(javaScript.get().replace("${downCurvatures}", JSON.toJSONString(downLineCurvatures)));
        //endregion

        //region 求拐点
//        TreeMap<Integer, Double[]> downTurnPoints = turnPointIndex(downLineCurvatures,downDenseFill);
//        TreeMap<Integer, Double[]> upTurnPoints = turnPointIndex(upLineCurvatures,upDenseFill);
//        javaScript.set(javaScript.get().replace("${upTurn}", JSON.toJSONString(upTurnPoints.values())));
//        javaScript.set(javaScript.get().replace("${downTurn}", JSON.toJSONString(downTurnPoints.values())));
        //endregion
        return Double.MIN_VALUE;
    }

    private double calDistance(Double[] p1, Double[] p2) {
        double dx = p1[0] - p2[0];
        double dy = p1[1] - p2[1];
        return Math.sqrt(dx * dx + dy * dy);
    }

    private Double[] calAveragePoint(List<Double[]> points) {
        Double[] p = new Double[]{0d, 0d};
        for (int i = 0; i < points.size(); i++) {
            Double[] p0 = points.get(i);
            p[0] = p[0] * i / (i + 1) + p0[0] / (i + 1);
            p[1] = p[1] * i / (i + 1) + p0[1] / (i + 1);
        }
        return p;
    }

    private List<Double[]> curvatureOptimize(LinkedList<Double[]> line) {

        int i = 1;
        do {
            Double[] p1 = line.get(i - 1);
            Double[] p2 = line.get(i);
            Double[] p3 = line.get(i + 1);
            List<Double[]> group = Arrays.asList(p1, p2, p3);
//            double v = calCurvature(group);
            if (Math.abs(cDeltaTheta(group)) < Math.PI / 18||calDistance(p1,p3)<150)
                line.remove(i);
            else i++;
        } while (i < line.size() - 1);
        List<Double[]> res = new ArrayList<>();
        List<Double[]> group = new ArrayList<>();
        for (int j = 0; j < line.size(); j++) {
        }
        return res;
    }

    private boolean judgeBoundary(Double[][] bgt, int start, int end, double k, double b, int direction) {

        int up = 0;
        int down = 0;
        int onLine = 0;
        for (int i = start; i * direction < end * direction; i += direction) {
            double temp = k * bgt[i][0] + b;
            temp = temp - bgt[i][1];
            if (Math.abs(temp) < 1e-8) temp = 0;
            if (temp > 0) up++;
            else if (temp < 0) down++;
            else onLine++;
            if (up > 0 && down > 0)
                return false;
        }
        return true;
    }

    public Double calSpe(Double[][] bgt, Double[][] rBGT, BGTParam param, int groupPointNum, double model) {

        int modelz = (int) model;
        if (rBGT == null) return null;

        TreeMap<Double, Integer> curUp = new TreeMap<>();//上行曲率变化
        TreeMap<Double, Integer> curDown = new TreeMap<>();//下行曲率变化
        TreeMap<Double, Integer> curDownTurn = new TreeMap<>(Comparator.reverseOrder());//下行拐点
        Double lastKi = null;
        Double lastKiC = null;
        for (int i = 0; i < rBGT.length / 2 - groupPointNum; i++) {
            List<Double[]> points = new ArrayList<>();
            for (int j = 0; j < groupPointNum; j++) points.add(rBGT[i + j]);
            double ki = calCurvature(points);
            if (lastKi == null) {
                lastKi = ki;
                continue;
            }
            curUp.put(ki - lastKi, groupPointNum / 2 + i - 1);

            lastKi = ki;
        }
        lastKi = null;
        for (int i = rBGT.length / 2; i < rBGT.length - groupPointNum; i++) {
            List<Double[]> points = new ArrayList<>();
            for (int j = 0; j < groupPointNum; j++) points.add(rBGT[i + j]);
            double ki = calCurvature(points);
            if (lastKi == null) {
                lastKi = ki;
                continue;
            }
            if (lastKiC != null && Math.abs(ki - lastKi - lastKiC) >= Math.abs(ki - lastKi + lastKiC)) {
//                System.out.println("turn point = " + bgt[groupPointNum / 2 + i - 1][0]);
//                System.out.println("index = " + (groupPointNum / 2 + i - 1) + "  ;ki =" + (ki - lastKi));
//                System.out.println("5 points:");
                for (int j = groupPointNum / 2 + i - 1; j > groupPointNum / 2 + i - 1 - groupPointNum; j--) {
                    if (bgt[j][1] <= (param.maxF - param.minF) / 3.5 + bgt[groupPointNum / 2 + i - 1][1])
                        curDownTurn.put(bgt[j][0], j);
//                    System.out.println("turn point i:{"+j+"} = " + bgt[j][0]);
                }
            }
//            System.out.println("index = " + (groupPointNum / 2 + i - 1) + "  ;ki =" + (ki - lastKi));
            curDown.put(ki - lastKi, groupPointNum / 2 + i - 1);
            lastKiC = ki - lastKi;

            lastKi = ki;
        }
//        System.out.println("curDown = " + curDown);
//        System.out.println("curUp = " + curUp);
//        System.out.println("curDownTurn = " + curDownTurn);
        switch (modelz) {
            case 1:
            case 2:
            case 3: {
                Map.Entry<Double, Integer> en1 = curUp.pollLastEntry();
                Map.Entry<Double, Integer> en2 = curDown.pollLastEntry();
                return Math.abs(bgt[en2.getValue()][0] - param.minS);
            }
            case 4: {
                Map.Entry<Double, Integer> en1 = curDown.pollLastEntry();
                Map.Entry<Double, Integer> en2 = curDown.pollLastEntry();
                return Math.abs((bgt[en1.getValue()][0] < bgt[en2.getValue()][0] ? bgt[en1.getValue()][0] : bgt[en2.getValue()][0]) - param.minS);
            }
            case 5:
            case 7: {
                Map.Entry<Double, Integer> en1 = curUp.pollLastEntry();
                Map.Entry<Double, Integer> en2 = curUp.pollLastEntry();
                return Math.abs(bgt[en1.getValue()][0] - bgt[en2.getValue()][0]);
            }
            case 6: {
                TreeMap<Double, Integer> s = new TreeMap<>();
                for (int i = 0; i < 3; i++) {
                    Map.Entry<Double, Integer> en = curDown.pollLastEntry();
                    s.put(bgt[en.getValue()][0], en.getValue());
                }
                Map.Entry<Double, Integer> en1 = s.pollFirstEntry();
                Map.Entry<Double, Integer> en2 = s.pollFirstEntry();
                return Math.abs(en1.getKey() - en2.getKey());
            }
            case 8: {
                Map.Entry<Double, Integer> en1 = curUp.pollLastEntry();
                Map.Entry<Double, Integer> en2 = curDown.pollLastEntry();
                double temp = bgt[en2.getValue()][0] - bgt[en1.getValue()][0];
                if (temp < 0) temp = 0;
                return temp;
            }
            case 9: {//中度供液不足
//                System.out.println("model = " + model);
                for (Double sTemp : curDownTurn.keySet()) {
                    if (sTemp / param.maxS <= 0.5) return sTemp - param.minS;
                }

                return curDownTurn.pollFirstEntry().getKey();
            }
            case 10: {//轻度供液不足
                for (Double sTemp : curDownTurn.keySet()) {
                    if (sTemp / param.maxS <= 0.75) return sTemp - param.minS;
                }
                return curDownTurn.pollFirstEntry().getKey();
            }
            case 11: {//重度供液不足
                for (Double sTemp : curDownTurn.keySet()) {
                    if (sTemp / param.maxS <= 0.25) return sTemp - param.minS;
                }
                return curDownTurn.pollFirstEntry().getKey();
            }
            default:
                return null;
        }
    }

    private double cSLoss(double spe, double maxS) {
        double temp = maxS - spe;
        if (temp < 0) return 0;
        else return temp;
    }

    /**
     * @param points
     * @return 绝对值
     */
    public double calCurvature(List<Double[]> points) {
        double deltaL = cDeltaL(points);
        double deltaTheta = cDeltaTheta(points);
        double v = Math.abs(deltaTheta / deltaL);
        return v < 1e-10 ? 0.0 : v;
    }

    private double cDeltaTheta(List<Double[]> points) {
        Double[] mid = points.get(points.size() / 2);
        Double[] first = points.get(0);
        Double[] last = points.get(points.size() - 1);
        double temp = (mid[1] - last[1]) * (mid[0] - first[0]) - (mid[1] - first[1]) * (mid[0] - last[0]);
        double temp2 = (mid[1] - last[1]) * (mid[1] - first[1]) + (mid[0] - first[0]) * (mid[0] - last[0]);
        return Math.atan(temp / temp2);
    }

    //备用计算方案
    /*  private double cDeltaTheta(List<Double[]> points) {
          Double[] mid = points.get(points.size() / 2);
          Double[] first = points.get(0);
          Double[] last = points.get(points.size() - 1);
         double theta1=Math.atan2(first[0]-mid[0],first[1]-mid[1]);
         double theta2=Math.atan2(last[0]-mid[0],last[1]-mid[1]);
         return Math.PI-theta1+theta2;
      }*/

    private double cDeltaL(List<Double[]> points) {
        double l = 0;
        for (int i = 0; i < points.size() - 1; i++) {
            Double[] p1 = points.get(i);
            Double[] p2 = points.get(i + 1);
            l += Math.sqrt((p1[0] - p2[0]) * (p1[0] - p2[0]) + (p1[1] - p2[1]) * (p1[1] - p2[1]));
        }
        return l;
    }

    static class BGTParam {
        public Double maxF;
        public Double minF;
        public Double maxS;
        public Double minS;


    }

    /**
     * 泵功图参数解析<br/>
     * <p>
     * 归一化功图 <br/>
     * <p>
     * 下冲程水平翻转与上冲程首尾相接组合为一条前向曲线
     *
     * @param bgt
     * @return
     */
    public Double[][] rBGT(Double[][] bgt, BGTParam param) {
        param.maxF = Double.MIN_VALUE;
        param.maxS = Double.MIN_VALUE;
        param.minF = Double.MAX_VALUE;
        param.minS = Double.MAX_VALUE;
        for (int i = 0; i < bgt.length; i++) {
            if (bgt[i][1] > param.maxF) param.maxF = bgt[i][1];
            if (bgt[i][0] > param.maxS) param.maxS = bgt[i][0];
            if (bgt[i][1] < param.minF) param.minF = bgt[i][1];
            if (bgt[i][0] < param.minS) param.minS = bgt[i][0];
        }
        if (param.maxF == Double.MIN_VALUE) return null;

//        boolean isDownStroke = false;
        Double[][] rBGT = new Double[bgt.length][2];
        for (int i = 0; i < bgt.length; i++) {
            rBGT[i][0] = bgt[i][0] / param.maxS;
            rBGT[i][1] = bgt[i][1] / param.maxF;
/*            if (i >= bgt.length / 2 && !isDownStroke)
                if (bgt[i][0] < maxS) isDownStroke = true;
            if (isDownStroke)
                rBGT[i][0] = 2 - rBGT[i][0];//此处会有几个点的误差，未必是下冲程
                */

            if (i >= bgt.length / 2) {
                rBGT[i][0] = 2 - rBGT[i][0];//此处会有几个点的误差，未必是下冲程
            }
        }

        return rBGT;
    }

    //endregion

    //region 计算供液不足类似功图有效冲程

    /**
     * 这个方向先暂停
     *
     * @param bgt
     * @param maxF
     * @param minF
     * @param maxS
     * @param minS
     * @return
     */
    @Deprecated
    public Double calSpe(Double[][] bgt, Double maxF, Double minF, Double maxS, Double minS) {

        double spe = maxS - minS;

        while (true) {
            double mid = minF + 0.01;
            TreeMap<Double, Integer> ps = new TreeMap<>();
            for (int i = bgt.length - 1; i >= bgt.length / 2; i--) {
                if (bgt[i][1] <= mid) {
                    ps.put(bgt[i][0], i);
                }
            }
            if (ps.size() <= 1) return 0d;
            minS = ps.pollFirstEntry().getKey();
            maxS = ps.pollLastEntry().getKey();
            if (Math.abs(maxS - minS - spe) < 0.01) return maxS - minS > spe ? maxS - minS : spe;
            else {
                spe = maxS - minS;
            }
        }
    }
    //endregion


    //region   液量计算

    /**
     * 理论排量计算
     *
     * @param s_pre
     * @param n
     * @param pumpDiameter
     * @return 体积排量
     */
    private double cTheoreticalDisplacement(double s_pre, double n, double pumpDiameter) {

        return s_pre * n * 24 * 60 * pumpDiameter * pumpDiameter * Math.PI / 4000000;
    }

    /**
     * 泵效计算
     *
     * @param s
     * @param spe
     * @return
     */
    private double cPumpEffect(double s, double spe) {
        if (StringUtils.isEmpty(s) || StringUtils.isEmpty(spe)) return 0;
        return spe / s * 100;
    }

    /**
     * @param s
     * @param spe
     * @return %
     */
    private double cStrokeRatio(double s, double spe) {
        if (s <= 0) return 0;
        return spe / s * 100;
    }

    /**
     * @param pumpDiameter
     * @param pumpDepth
     * @param waterCut      % 原油体积含水率
     * @param oilDensity    0.85 地面原油比重
     * @param waterDensity  1.05  采出水密度
     * @param n
     * @param maxS
     * @param maxF
     * @param minF
     * @param isPumpAndGush
     * @return
     */
    private double cProductionCoefficient(double pumpDiameter, double pumpDepth, double waterCut,
                                          double oilDensity, double waterDensity,
                                          double n, double maxS, double maxF,
                                          double minF, boolean isPumpAndGush) {
        double L_TheoryLoadLiquid = Math.PI / 4 * (pumpDiameter * pumpDiameter / 1000000) * pumpDepth *
                ((1 - waterCut / 100) * oilDensity + waterCut / 100 * waterDensity) * 9.8 *
                (1 + n * maxS / 127);
        if ((maxF - minF) / L_TheoryLoadLiquid < 0.8 && !isPumpAndGush &&
                pumpDepth > 1750)
            return (maxF - minF) / L_TheoryLoadLiquid;
        else return 1;

    }

    private double cLiqProdDailyVolume(double pumpDiameter, double n, double s, double strokeRatio, double pumpDepth, double Bo) {
//        if (Bo < 1) Bo = 1;//体积系数必须大于1 放到单井特性里去做
        double v = pumpDiameter * pumpDiameter * Math.PI / 4 * 24 * 60 * n * s * strokeRatio / 100000000;
        if (v > 0.0087 * Math.pow(2.71828, 0.0027 * pumpDepth))
            return v * Bo;//---0.0087*power(2.71828,0.0027*nvl(L_Bs,0));
        else return v * Bo;
    }

    /**
     * @param liqProdDailyVolume
     * @param waterCut           % 原油体积含水率
     * @param gasOilRatio        气油比 体积比
     * @return
     */
    private double cGasProdDaily(double liqProdDailyVolume, double waterCut, double gasOilRatio) {
        return gasOilRatio * (liqProdDailyVolume - liqProdDailyVolume * waterCut / 100);
    }

    /**
     * @param liqProdDailyVolume
     * @param waterCut              % 原油体积含水率
     * @param productionCoefficient
     * @param oilDensity
     * @return
     */
    private double cOilProdDaily(double liqProdDailyVolume, double waterCut, double productionCoefficient, double oilDensity) {
        return oilDensity * (liqProdDailyVolume - liqProdDailyVolume * waterCut / 100) * productionCoefficient;
    }

    /**
     * @param liqProdDailyVolume
     * @param waterCut              % 原油体积含水率
     * @param productionCoefficient
     * @param waterDensity
     * @return
     */
    private double cWaterProdDaily(double liqProdDailyVolume, double waterCut, double productionCoefficient, double waterDensity) {
        return waterDensity * liqProdDailyVolume * waterCut / 100 * productionCoefficient;
    }

    private double cLiqProdDaily(double waterProdDaily, double oilProdDaily) {
        return waterProdDaily + oilProdDaily;
    }

    private double cDynamicLevel(double oilDensity, double waterCut, double waterDensity, double pumpDiameter, double s, double spe, double minF, double pumpDepth) {
        double temp = ((1 - waterCut / 100) * oilDensity + waterCut / 100 * waterDensity) * 9.8 * 0.25 * Math.PI * Math.pow(pumpDiameter, 2) / 1000;
        double submergenceDepth = 0;

        if (temp == 0)
            return pumpDepth - submergenceDepth;
        else {
            submergenceDepth = Math.abs(minF * spe / s) * 1000 / temp;
            return pumpDepth - submergenceDepth;
        }

    }
    //endregion

    public CalProductionResult calProduction(Double[][] bgt, double s,
                                             double n, boolean isPumpAndGush,
                                             double pumpDiameter, double pumpDepth,
                                             double waterCut, double waterDensity,
                                             double oilDensity, double gasOilRatio,
                                             double Bo, double model) {


        BGTParam param = new BGTParam();
        Double[][] rBGT = rBGT(bgt, param);
        if (s < param.maxS) s = param.maxS;
        Double spe = calSpe(bgt, rBGT, param, 5, model);
        CalProductionResult result = new CalProductionResult();
        result.bx =cPumpEffect(s, spe);
        result.l_StrokeRatio = cPumpEffect(s, spe);
        result.lsss = cSLoss(spe, s);
        result.llpl = cTheoreticalDisplacement(s, n, pumpDiameter);
        double productionCoefficient = cProductionCoefficient(pumpDiameter, pumpDepth, waterCut, oilDensity, waterDensity, n, param.maxS, param.maxF, param.minF, isPumpAndGush);
        double v = cLiqProdDailyVolume(pumpDiameter, n, s, result.l_StrokeRatio, pumpDepth, Bo);
        result.yxcc = spe;
        result.rcql = cGasProdDaily(v, waterCut, gasOilRatio);
        result.rcyl1 = cOilProdDaily(v, waterCut, productionCoefficient, oilDensity);
        result.rcsl = cWaterProdDaily(v, waterCut, productionCoefficient, waterDensity);
        result.rcyl = cLiqProdDaily(result.rcsl, result.rcyl1);
        result.dym = cDynamicLevel(oilDensity, waterCut, waterDensity, pumpDiameter, s, spe, param.minF, pumpDepth);
        result.formatted();
        return result;
    }
}
